/* Copyright (C) 1995 Free Software Foundation, Inc.
This file is part of the GNU C Library.

The GNU C Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU C Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU C Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

__MATH_INLINE double __hypot(const double x, const double y)
{
	register double argx = x, argy = y;

	return sqrt(argx*argx+argy*argy);
}

__MATH_INLINE double __log1p(const double x)
{
	register double value, arg = x;

	if (fabs (arg) >= 1.0-0.5*M_SQRT2)
		value = log(1.0+arg);

	else {
		__asm__ __volatile__
			("fldln2\n\t"
			 "fxch\n\t"
			 "fyl2xp1"
			 : "=t" (value) : "0" (arg));
	}

	return value;
}

__MATH_INLINE double __asinh(const double x)
{
	register double arg = x, y=fabs(arg);
	
	return __log1p(y*y/(sqrt(y*y+1.0)+1.0)+y)*__sgn1(arg);
}

__MATH_INLINE double __acosh(const double x)
{
	register double arg = x;

	return log(arg+sqrt(arg-1.0)*sqrt(arg+1.0));
}

__MATH_INLINE double __atanh(const double x)
{
	register double arg = x, y=fabs(arg);

	return -0.5*__log1p(-(y+y)/(1.0+y))*__sgn1(arg);
}

__MATH_INLINE double __coshm1(const double x)
{
	register double exm1 = __expm1(fabs(x));

	return 0.5*(exm1/(exm1+1.0))*exm1;
}

__MATH_INLINE double __acosh1p(const double x)
{
	register double arg = x;

	return __log1p(arg+sqrt(arg)*sqrt(arg+2.0));
}

__MATH_INLINE double __logb(const double x)
{
	register double value, arg = x, dummy;
	__asm__ __volatile__
		("fxtract\n\t"
		 : "=t" (dummy), "=u" (value): "0" (arg));

	return	value;
}

__MATH_INLINE double __drem(const double x, const double y)
{
	register double value, value2, argx = x, argy = y;
	__asm__ __volatile__
		("1:	fprem1\n\t"
		 "fstsw	%%ax\n\t"
		 "sahf\n\t"
		 "jp	1b" : "=t" (value), "=u" (value2)
		 : "0" (argx), "1" (argy) : "ax", "cc");

	return value;
}

__MATH_INLINE void __sincos(const double x, double *sinx, double *cosx)
{
	register double cosr, sinr, arg = x;
	__asm__ __volatile__
		("fsincos": "=t" (cosr), "=u" (sinr) : "0" (arg));

	*sinx = sinr; *cosx = cosr;
}

__MATH_INLINE double __sgn(const double x)
{
	return (x == 0.0 ? 0.0 : (x > 0.0 ? 1.0 : -1.0));
}

__MATH_INLINE double __pow2(const double x)
{
	register double value, arg = x, exponent;
	long p=(long) x;
	
	if(arg == (double) p) return ldexp(1.0,p);
	__asm__ __volatile__
		("fldl	%%st(0)\n\t"
		 "frndint				# int(x)\n\t"
		 "fxch\n\t"
		 "fsub	%%st(1)			# fract(x)\n\t"
		 "f2xm1				# 2^(fract(x)) - 1\n\t"
		 : "=t" (value), "=u" (exponent) : "0" (arg));
	value += 1.0;
	__asm__ __volatile__
		("fscale" : "=t" (value): "0" (value), "u" (exponent));
	return value;
}


#ifndef __STRICT_ANSI__

#if __OPTIMIZE__

#define expm1 __expm1
#define hypot __hypot
#define log1p __log1p
#define asinh __asinh
#define acosh __acosh
#define atanh __atanh
#define coshm1 __coshm1
#define acosh1p __acosh1p
#define drem __drem
#define sincos __sincos
#define sgn __sgn
#define sgn1 __sgn1
#define log2 __log2
#define pow2 __pow2

#else

#define sincos(x,sinx,cosx) (*sinx=sin(x), *cosx=cos(x))
#define sgn(x) ((x) == 0.0 ? 0.0 : ((x) > 0.0 ? 1.0 : -1.0))
#define sgn1(x) ((x) >= 0.0 ? 1.0 : -1.0)
#define log2(x) (M_LOG2E*log(x));

#endif /* __OPTIMIZE__ */

#endif /* __STRICT_ANSI__ */
